#!/bin/python
#-----------------------------------------------------------------------------
# File Name : create_hdf5.py
# Author: Emre Neftci
#
# Creation Date : Tue Nov  5 13:15:54 2019
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv2
#----------------------------------------------------------------------------- 
import numpy as np
from tqdm import tqdm
import scipy.misc
import h5py
import glob
import torch.utils.data
from ..events_timeslices import *
from ..utils import *
import os

def gather_aedat(directory, filename_prefix = ''):
    if not os.path.isdir(directory):
        raise FileNotFoundError("aedat files not found, looked at: {}".format(directory))
    import glob
    fns = []
    search_mask = directory+'/'+filename_prefix+'*.aedat'
    glob_out = glob.glob(search_mask)
    if len(glob_out)>0:
        fns+=glob_out
    return fns


def create_events_hdf5(directory, hdf5_filename):
    fns = gather_aedat(directory)

    keys = []
    with h5py.File(hdf5_filename, 'w') as f:
        f.clear()

        key = 0
        metas = []
        data_grp = f.create_group('data')
        extra_grp = f.create_group('extra')
        for file_d in tqdm(fns):
            istrain = file_d in fns
            data, labels_starttime = aedat_to_events(file_d)
            tms = data[:,0]
            ads = data[:,1:]
            #lbls = labels_starttime[:,0]
            out = []

            keys.append(key)
            s_ = get_slice(tms, ads, start_tms[i], end_tms[i])
            times = s_[0]
            addrs = s_[1]
            metas.append({}) 
            subgrp = data_grp.create_group(str(key))
            tm_dset = subgrp.create_dataset('times' , data=times, dtype=np.uint32)
            ad_dset = subgrp.create_dataset('addrs' , data=addrs, dtype=np.uint8)
            #lbl_dset= subgrp.create_dataset('labels', data=lbls[i]-1, dtype=np.uint8)
            subgrp.attrs['meta_info']= str(metas[-1])
            key += 1
        extra_grp.create_dataset('keys', data=train_keys)
        extra_grp.attrs['N'] = len(keys) 
            


if __name__=="__main__":
    out = gather_aedat('./')






#!/bin/python
#-----------------------------------------------------------------------------
# File Name : neuromorphic_dataset.py
# Author: Emre Neftci
#
# Creation Date : Thu Nov  7 11:50:48 2019
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv2
#----------------------------------------------------------------------------- 
import os
import torch
import torch.utils.data as data
from torchvision.datasets.utils import extract_archive, verify_str_arg, check_integrity, gen_bar_updater
from .transforms import Compose

DEFAULT_ROOT = 'data/'

def download_url(url, root, filename=None, md5=None):
    """Download a file from a url and place it in root.
    Args:
        url (str): URL to download file from
        root (str): Directory to place downloaded file in
        filename (str, optional): Name to save the file under. If None, use the basename of the URL
        md5 (str, optional): MD5 checksum of the download. If None, do not check
    """
    from six.moves import urllib

    root = os.path.expanduser(root)
    if not filename:
        filename = os.path.basename(url)
    fpath = os.path.join(root, filename)


    os.makedirs(root, exist_ok=True)

    # check if file is already present locally
    if check_integrity(fpath, md5):
        print('Using downloaded and verified file: ' + fpath)
    else:   # download the file
        try:
            if 'dropbox' in url:
                # Handle dropbox links differently
                import requests
                headers = {'user-agent': 'Wget/1.16 (linux-gnu)'}
                r = requests.get(url, stream=True, headers=headers)
                with open(fpath, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=1024): 
                        if chunk:
                            f.write(chunk)
            elif 'Manual' in url:
                raise urllib.error.URLError(url)
            else:
                print('Downloading ' + url + ' to ' + fpath)
                urllib.request.urlretrieve(
                    url, fpath,
                    reporthook=gen_bar_updater()
                )
        except (urllib.error.URLError, IOError) as e:
            if url[:5] == 'https':
                url = url.replace('https:', 'http:')
                print('Failed download. Trying https -> http instead.'
                      ' Downloading ' + url + ' to ' + fpath)
                urllib.request.urlretrieve(
                    url, fpath,
                    reporthook=gen_bar_updater()
                )
            else:
                raise e
        # check integrity of downloaded file
        if not check_integrity(fpath, md5):
            raise RuntimeError("File not found or corrupted.")

def download_and_extract_archive(url, download_root, extract_root=None, filename=None,
                                 md5=None, remove_finished=False):
    download_root = os.path.expanduser(download_root)
    if extract_root is None:
        extract_root = download_root
    if not filename:
        filename = os.path.basename(url)

    download_url(url, download_root, filename, md5)

    archive = os.path.join(download_root, filename)
    print("Extracting {} to {}".format(archive, extract_root))
    extract_archive(archive, extract_root, remove_finished)

def identity(x):
    return x

class NeuromorphicDataset(data.Dataset):
    _repr_indent = 4

    def __init__(self, root=None, transforms = None, transform=None, target_transform=None):
        if isinstance(root, torch._six.string_classes):
            root = os.path.expanduser(root)
        self.root = root
 
        if root is not None:
            if not os.path.isfile(root):
                if self.download_and_create:
                    self.download()
                    self.create_hdf5()
                else:
                    raise Exception("File {} does not exist and download_and_create is False".format(root))




        has_transforms = transforms is not None
        has_separate_transform = transform is not None or target_transform is not None
        if has_transforms and has_separate_transform:
            raise ValueError("Only transforms or transform/target_transform can "
                             "be passed as argument")

        # for backwards-compatibility
        if transform is None:
            transform = identity
        if target_transform is None:
            target_transform = identity

        self.transform = transform
        self.target_transform = target_transform

        if has_separate_transform:
            transforms = StandardTransform(transform, target_transform)
        self.transforms = transforms

    def __getitem__(self, index):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __repr__(self):
        head = "Dataset " + self.__class__.__name__
        body = ["Number of datapoints: {}".format(self.__len__())]
        if self.root is not None:
            body.append("Root location: {}".format(self.root))
        body += self.extra_repr().splitlines()
        if hasattr(self, "transforms") and self.transforms is not None:
            body += [repr(self.transforms)]
        lines = [head] + [" " * self._repr_indent + line for line in body]
        return '\n'.join(lines)

    def _format_transform_repr(self, transform, head):
        lines = transform.__repr__().splitlines()
        return (["{}{}".format(head, lines[0])] +
                ["{}{}".format(" " * len(head), line) for line in lines[1:]])

    def extra_repr(self):
        return ""

    def check_exists(self):
        res_ = [os.path.exists(d) for d in self.resources_local]
        res = all(res_)
        if res is False:
            print('The following files did not exist, will attempt download:')
            for i,r in enumerate(res_):
                if not r: print(self.resources_local[i])
        return res


    def download(self):
        if self.check_exists():
            return True
        else:
            os.makedirs(self.directory, exist_ok=True)
            for url, md5, filename in self.resources_url:
                download_and_extract_archive(url, download_root=self.directory, filename=filename, md5=md5)
            return False

    def create_hdf5(self):
        raise NotImplementedError()

    def target_transform_append(self, transform):
        if transform is None:
            return
        if self.target_transform is None:
            self.target_transform = transform
        else:
            self.target_transform = Compose([self.target_transform, transform])


    def transform_append(self, transform):
        if transform is None:
            return
        if self.transform is None:
            self.transform = transform
        else:
            self.transform = Compose([self.transform, transform])




class StandardTransform(object):
    def __init__(self, transform=None, target_transform=None):
        self.transform = transform
        self.target_transform = target_transform

    def __call__(self, input, target):
        if self.transform is not None:
            input = self.transform(input)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return input, target

    def _format_transform_repr(self, transform, head):
        lines = transform.__repr__().splitlines()
        return (["{}{}".format(head, lines[0])] +
                ["{}{}".format(" " * len(head), line) for line in lines[1:]])

    def __repr__(self):
        body = [self.__class__.__name__]
        if self.transform is not None:
            body += self._format_transform_repr(self.transform,
                                                "Transform: ")
        if self.target_transform is not None:
            body += self._format_transform_repr(self.target_transform,
                                                "Target transform: ")

        return '\n'.join(body)



#!/bin/python
#-----------------------------------------------------------------------------
# File Name :
# Author: Emre Neftci
#
# Creation Date : Tue Nov  5 13:16:48 2019
# Last Modified :
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv2
#-----------------------------------------------------------------------------
import struct
import numpy as np

# adapted from https://github.com/gorchard/event-Python/blob/master/eventvision.py
def load_ATIS_bin(filename):
    """Reads in the TD events contained in the N-MNIST/N-CALTECH101 dataset file specified by 'filename'"""
    f = open(filename, 'rb')
    raw_data = np.fromfile(f, dtype=np.uint8)
    f.close()
    raw_data = np.uint32(raw_data)

    all_y = raw_data[1::5]
    all_x = raw_data[0::5]
    all_p = (raw_data[2::5] & 128) >> 7 #bit 7
    all_ts = ((raw_data[2::5] & 127) << 16) | (raw_data[3::5] << 8) | (raw_data[4::5])

    #Process time stamp overflow events
    time_increment = 2 ** 13
    overflow_indices = np.where(all_y == 240)[0]
    for overflow_index in overflow_indices:
        all_ts[overflow_index:] += time_increment

    #Everything else is a proper td spike
    td_indices = np.where(all_y != 240)[0]
    return all_ts, all_x, all_y, all_p

def load_jaer(datafile='/tmp/aerout.dat', length=0, version='aedat', debug=1, camera='DVS128'):
    """
    load AER data file and parse these properties of AE events:
    - timestamps (in us),
    - x,y-position [0..127]
    - polarity (0/1)

    @param datafile - path to the file to read
    @param length - how many bytes(B) should be read; default 0=whole file
    @param version - which file format version is used: "aedat" = v2, "dat" = v1 (old)
    @param debug - 0 = silent, 1 (default) = print summary, >=2 = print all debug
    @param camera='DVS128' or 'DAVIS240'
    @return (ts, xpos, ypos, pol) 4-tuple of lists containing data of all events;
    """
    # constants
    aeLen = 8  # 1 AE event takes 8 bytes
    readMode = '>II'  # struct.unpack(), 2x ulong, 4B+4B
    td = 0.000001  # timestep is 1us
    if (camera == 'DVS128'):
        xmask = 0x00fe
        xshift = 1
        ymask = 0x7f00
        yshift = 8
        pmask = 0x1
        pshift = 0
    elif (camera == 'DAVIS240'):  # values take from scripts/matlab/getDVS*.m
        xmask = 0x003ff000
        xshift = 12
        ymask = 0x7fc00000
        yshift = 22
        pmask = 0x800
        pshift = 11
        eventtypeshift = 31
    else:
        raise ValueError("Unsupported camera: %s" % (camera))

    if (version == 'dat'):
        print ("using the old .dat format")
        aeLen = 6
        readMode = '>HI'  # ushot, ulong = 2B+4B

    aerdatafh = open(datafile, 'rb')
    k = 0  # line number
    p = 0  # pointer, position on bytes
    statinfo = os.stat(datafile)
    if length == 0:
        length = statinfo.st_size

    # header
    lt = aerdatafh.readline()
    while lt and lt[0] == '#':
        p += len(lt)
        k += 1
        lt = aerdatafh.readline()
        if debug >= 2:
            print (str(lt))
        continue

    # variables to parse
    timestamps = []
    xaddr = []
    yaddr = []
    pol = []

    # read data-part of file
    aerdatafh.seek(p)
    s = aerdatafh.read(aeLen)
    p += aeLen

    # print (xmask, xshift, ymask, yshift, pmask, pshift)
    while p < length:
        addr, ts = struct.unpack(readMode, s)
        # parse event type
        if (camera == 'DAVIS240'):
            eventtype = (addr >> eventtypeshift)
        else:  # DVS128
            eventtype = 0

        # parse event's data
        if (eventtype == 0):  # this is a DVS event
            x_addr = (addr & xmask) >> xshift
            y_addr = (addr & ymask) >> yshift
            a_pol = (addr & pmask) >> pshift

            if debug >= 3:
                print("ts->", ts)  # ok
                print("x-> ", x_addr)
                print("y-> ", y_addr)
                print("pol->", a_pol)

            timestamps.append(ts)
            xaddr.append(x_addr)
            yaddr.append(y_addr)
            pol.append(a_pol)

        aerdatafh.seek(p)
        s = aerdatafh.read(aeLen)
        p += aeLen

    if debug > 0:
        try:
            print ("read %i (~ %.2fM) AE events, duration= %.2fs" % (
                len(timestamps), len(timestamps) / float(10 ** 6), (timestamps[-1] - timestamps[0]) * td))
            n = 5
            print ("showing first %i:" % (n))
            print ("timestamps: %s \nX-addr: %s\nY-addr: %s\npolarity: %s" % (
                timestamps[0:n], xaddr[0:n], yaddr[0:n], pol[0:n]))
        except:
            print ("failed to print statistics")

    return np.array(timestamps), np.array(xaddr), np.array(yaddr), np.array(pol)




def plot_frames_imshow(images, labels=None, nim=11, avg=50, interval=1, do1h = True, transpose=False, label_mapping=None):
    from matplotlib.colors import ListedColormap, LinearSegmentedColormap
    from matplotlib.pyplot import Normalize
    colors = ['red', 'black', 'green']
    cmap = LinearSegmentedColormap.from_list('name', colors)
    if avg>images.shape[1]:
        avg = images.shape[1] 

    rnge = range(0,np.maximum(images.shape[1]//avg,1),interval)

    import pylab as plt
    plt.figure(figsize = [nim+2,16])
    import matplotlib.gridspec as gridspec
    if not transpose:
        gs = gridspec.GridSpec(len(rnge), nim)
    else:
        gs = gridspec.GridSpec(nim, len(rnge))
    plt.subplots_adjust(left=0, bottom=0, right=1, top=0.95, wspace=.0, hspace=.04)
    if labels is not None:
        if do1h:
            categories = labels.argmax(axis=1)
        else:
            categories = labels
    else:
        categories = range(len(images))
    s=[]
    for j in range(nim):
         norm = Normalize(-images[j].mean()*30,images[j].mean()*30)
         for e,i in enumerate(rnge):
             if not transpose:
                 ax = plt.subplot(gs[e, j])
             else:
                 ax = plt.subplot(gs[j, e])
             plt.imshow(images[j,i*avg:(i*avg+avg),0,:,:].mean(axis=0).T -images[j,i*avg:(i*avg+avg),1,:,:].mean(axis=0).T, cmap=cmap, norm= norm)
             plt.xticks([])

             if i==0 and label_mapping is not None:
                 plt.title(label_mapping[int(categories[j])], fontsize=10)
             plt.yticks([])
         s.append(images[j].sum())

def legacy_aedat_to_events(filename, normalize_time = True):
    '''
    Uses the dv package to extract events from aedat 2 and aedat 3
    '''
    from dv import LegacyAedatFile
    events=[]

    with LegacyAedatFile(filename) as f:
        for event in f:
            events.append([event.timestamp,event.polarity,event.x,event.y])

    events = np.column_stack(np.array(events, dtype='uint32')).T
    if normalize_time:
        events[:,0] -= events[0,0]

    return events


def aedat_to_events(filename):
    '''
    Used for aedat 3.1
    '''
    label_filename = filename[:-6] +'_labels.csv'
    labels = np.loadtxt(label_filename, skiprows=1, delimiter=',',dtype='uint32')
    events=[]
    with open(filename, 'rb') as f:
        for i in range(5):
            f.readline()
        while True:
            data_ev_head = f.read(28)
            if len(data_ev_head)==0: break

            eventtype = struct.unpack('H', data_ev_head[0:2])[0]
            eventsource = struct.unpack('H', data_ev_head[2:4])[0]
            eventsize = struct.unpack('I', data_ev_head[4:8])[0]
            eventoffset = struct.unpack('I', data_ev_head[8:12])[0]
            eventtsoverflow = struct.unpack('I', data_ev_head[12:16])[0]
            eventcapacity = struct.unpack('I', data_ev_head[16:20])[0]
            eventnumber = struct.unpack('I', data_ev_head[20:24])[0]
            eventvalid = struct.unpack('I', data_ev_head[24:28])[0]

            if(eventtype == 1):
                event_bytes = np.frombuffer(f.read(eventnumber*eventsize), 'uint32')
                event_bytes = event_bytes.reshape(-1,2)

                x = (event_bytes[:,0] >> 17) & 0x00001FFF
                y = (event_bytes[:,0] >> 2 ) & 0x00001FFF
                p = (event_bytes[:,0] >> 1 ) & 0x00000001
                t = event_bytes[:,1]
                events.append([t,x,y,p])

            else:
                f.read(eventnumber*eventsize)
    events = np.column_stack(events)
    events = events.astype('uint32')
    clipped_events = np.zeros([4,0],'uint32')
    for l in labels:
        start = np.searchsorted(events[0,:], l[1])
        end = np.searchsorted(events[0,:], l[2])
        clipped_events = np.column_stack([clipped_events,events[:,start:end]])
    return clipped_events.T, labels

def rosbag_to_events(filename, topic='/dvs_right/events'):
    try:
        from importRosbag.importRosbag import importRosbag
    except ImportError as exc:
        print("This function requires the importRosbag library from https://github.com/event-driven-robotics")
        raise(exc)
    all_events = []
    data = importRosbag(filename)[topic]
    data['ts'] -= data['ts'][0] # align at 0
    data['ts'] *= 1000000. # second to microsecond
    return data



#!/bin/python
#-----------------------------------------------------------------------------
# File Name : test_dvsgestures.py
# Author: Emre Neftci
#
# Creation Date : Fri 19 Sep 2019 
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv3
#----------------------------------------------------------------------------- 
from torchneuromorphic.nmnist.nmnist_dataloaders import *
from torchneuromorphic.utils import plot_frames_imshow
import torchneuromorphic.transforms as transforms
from torch.utils.data import Subset, SubsetRandomSampler

def create_dataloader(
        root = 'data/nmnist/n_mnist.hdf5',
        batch_size = 72 ,
        chunk_size_train = 300,
        chunk_size_test = 300,
        dt = 1000,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None,
        **dl_kwargs):
    
    ds = [1,1] 
    low_crop = [0,0]
    high_crop = [32,32]
    size = [2, np.ceil((high_crop[0]-low_crop[0])/ds[0]).astype('int'), np.ceil((high_crop[1]-low_crop[1])/ds[1]).astype('int')]

    print(size)
    default_transform = lambda chunk_size:  transforms.Compose([
         transforms.CropDims(low_crop,high_crop,[2,3]),
         transforms.Downsample(factor=[dt,1,ds[0],ds[1]]),
         transforms.ToCountFrame(T = chunk_size, size = size),
         transforms.ToTensor()
    ])


    if transform_train is None:
        transform_train = default_transform(chunk_size_train)
    if transform_test is None:
        transform_test = default_transform(chunk_size_test)

    if target_transform_train is None:
        target_transform_train = transforms.Compose([transforms.toDtype(torch.long)])
    if target_transform_test is None:
        target_transform_test =  transforms.Compose([transforms.toDtype(torch.long)])

    train_d = NMNISTDataset(root,
                            train=True,
                            transform = transform_train, 
                            target_transform = target_transform_train, 
                            chunk_size = chunk_size_train)


    train_subset_indices = train_d.keys_by_label[:,:100].reshape(-1)

    train_dl = torch.utils.data.DataLoader(train_d, batch_size=batch_size, sampler=SubsetRandomSampler(train_subset_indices), **dl_kwargs)

    test_d = NMNISTDataset(root,
                               transform = transform_test, 
                               target_transform = target_transform_test, 
                               train=False,
                               chunk_size = chunk_size_test,
                               )


    test_subset_indices = test_d.keys_by_label[:,:100].reshape(-1) 
    test_dl = torch.utils.data.DataLoader(test_d, batch_size=batch_size, sampler=SubsetRandomSampler(test_subset_indices), **dl_kwargs)
    return train_dl, test_dl

if __name__ == "__main__":
    train_dl, test_dl = create_dataloader()

    import tqdm
    for d,t in tqdm.tqdm(iter(test_dl)): 
        print(d.shape)



#!/usr/bin/env python
#-----------------------------------------------------------------------------
# Author: Jacques Kaiser
#
# Creation Date : Wed 13 May 2020
#
# Copyright : (c)
# Licence : Apache License, Version 2.0
#-----------------------------------------------------------------------------
import h5py
import torch.utils.data
import numpy as np
import glob
import os

from .create_hdf5 import create_events_hdf5
from ..neuromorphic_dataset import NeuromorphicDataset
from ..events_timeslices import *
from ..transforms import *

class RosbagDataset(NeuromorphicDataset):
    def __init__(
            self,
            root,
            resources_local=None,
            train=True,
            transform=None,
            target_transform=None,
            download_and_create=True,
            chunk_size = 500):

        self.directory = 'data/rosbags/'
        if resources_local is None:
            resources_local = [os.path.join(self.directory, 'raw')]
        self.resources_local = resources_local
        no_rosbag_found = 'No rosbag found. Place your rosbags in {}/{}'.format(self.directory, self.resources_local[0])
        self.resources_url = [[no_rosbag_found, None, 'your rosbags']]

        self.n = 0
        self.download_and_create = download_and_create
        self.root = root
        self.train = train
        self.chunk_size = chunk_size

        super(RosbagDataset, self).__init__(
                root,
                transform=transform,
                target_transform=target_transform )

        with h5py.File(root, 'r', swmr=True, libver="latest") as f:
            self.label_order = f['extra']['label_order'][()]
            print("Labels in order: {}".format(self.label_order))
            self.n_labels = len(self.label_order)
            if train:
                self.n = f['extra'].attrs['Ntrain']
                self.keys = f['extra']['train_keys'][()]
            else:
                self.n = f['extra'].attrs['Ntest']
                self.keys = f['extra']['test_keys'][()]

    def download(self):
        super(RosbagDataset, self).download()

    def create_hdf5(self):
        create_events_hdf5(self.resources_local[0], self.root)

    def __len__(self):
        return self.n

    def __getitem__(self, key):
        #Important to open and close in getitem to enable num_workers>0
        with h5py.File(self.root, 'r', swmr=True, libver="latest") as f:
            if not self.train:
                key = key + f['extra'].attrs['Ntrain']
            assert key in self.keys
            data, target = sample(
                    f,
                    key,
                    T = self.chunk_size,
                    shuffle=self.train)

        if self.transform is not None:
            data = self.transform(data)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return data, target


def sample(hdf5_file,
        key,
        T = 500,
        shuffle = False):
    dset = hdf5_file['data'][str(key)]
    label = dset['labels'][()]
    tend = dset['times'][-1]
    start_time = 0
    tmad = get_tmad_slice(dset['times'][()], dset['addrs'][()], start_time, T*1000)
    tmad[:,0]-=tmad[0,0]
    return tmad, label


def create_dataloader(
        root = 'data/rosbags/rosbags_build.hdf5',
        batch_size = 72 ,
        chunk_size_train = 500,
        chunk_size_test = 1800,
        ds = None,
        dt = 1000,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None,
        n_events_attention=None,
        **dl_kwargs):
    if ds is None:
        ds = 4
    size = [2, 128//ds, 128//ds]
    center = [64,64]

    default_transform = lambda chunk_size: Compose([
        CropCenter(center, size),
        Downsample(factor=[dt,1,1,1]),
        ToCountFrame(T = chunk_size, size = size),
        ToTensor()
    ])

    if transform_train is None:
        transform_train = default_transform(chunk_size_train)
    if transform_test is None:
        transform_test = default_transform(chunk_size_test)

    if target_transform_train is None:
        target_transform_train =Compose([Repeat(chunk_size_train), toOneHot(4)]) # HACK HACK HACK! Can't get n_labels from here
    if target_transform_test is None:
        target_transform_test = Compose([Repeat(chunk_size_test), toOneHot(4)])

    train_d = RosbagDataset(root,
                            train=True,
                            transform = transform_train,
                            target_transform = target_transform_train,
                            chunk_size = chunk_size_train)

    train_dl = torch.utils.data.DataLoader(train_d, batch_size=batch_size, shuffle=True, **dl_kwargs)

    test_d = RosbagDataset(root,
                           transform = transform_test,
                           target_transform = target_transform_test,
                           train=False,
                           chunk_size = chunk_size_test)

    test_dl = torch.utils.data.DataLoader(test_d, batch_size=batch_size, **dl_kwargs)

    return train_dl, test_dl



#!/bin/python
#-----------------------------------------------------------------------------
# File Name : double_nmnist_dataloaders.py
# Author: Emre Neftci
#
# Creation Date : Tue 26 May 2020 02:25:41 PM PDT
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : Apache License, Version 2.0
#----------------------------------------------------------------------------- 

import struct
import time
import numpy as np
import scipy.misc
import h5py
import torch.utils.data
from .neuromorphic_dataset import NeuromorphicDataset 
from .events_timeslices import *
from .transforms import *

import os


class CombinedDataset(NeuromorphicDataset):
    def __init__(self,
                base_datasets,
                train = True,
                transform = None,
                target_transform = None,
                chunk_size = 500):

        self.download_and_create   = False
        super(CombinedDataset, self).__init__(
        root = None,
        transform=transform,
        target_transform=target_transform )

        self.base_datasets = d = base_datasets
        self.nd = len(base_datasets)
        self.lens = lens = [len(d) for d in base_datasets]
        self.lensmd = lensmd = np.cumsum(lens[:-1], dtype=int)
    
        self.keys = lambda d: [d//l for l in lensmd]+[d%lens[0]]
        self.locks = lambda ts: sum(ts*np.array([d[i].nclasses for i in range(1,len(d))]+[1]))
 
    def __len__(self):
        return np.prod(self.lens,0)


    def __getitem__(self, key):
        #Important to open and close in getitem to enable num_workers>0
        keys_ = self.keys(key)
        d = self.base_datasets
        datas, targets = list(zip(*[d[i][k] for i,k in enumerate(keys_)] ))
        data = sum(datas)
        target = self.locks([t[-1].argmax(0) for i,t in enumerate(targets)])
        
        if self.transform is not None:
            data = self.transform(data)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return data, target




#!/bin/python
#-----------------------------------------------------------------------------
# File Name : create_dvsgestures.py
# Author: Emre Neftci
#
# Creation Date : Tue Nov  5 13:20:05 2019
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv2
#----------------------------------------------------------------------------- 
from torchneuromorphic.dvssign.dvssign_dataloaders import *

if __name__ == "__main__":
    DVSSignDataset = DVSSignDataset(root='/home/kennetms/Documents/torchneuromorphic/torchneuromorphic/data/ASL-DVS/dvssign.hdf5')
    
    #out = create_events_hdf5('/home/kennetms/Documents/torchneuromorphic/torchneuromorphic/data/ASL-DVS', '/home/kennetms/Documents/torchneuromorphic/torchneuromorphic/data/ASL-DVS/dvssign.hdf5')

    f = h5py.File('/home/kennetms/Documents/torchneuromorphic/torchneuromorphic/data/ASL-DVS/dvssign.hdf5', 'r')
    
    print(list(f.keys()))
    
    data = f['data']
    
    print(data['0'].keys())
    
    print("addrs",data['0']['addrs'][0])
    
    print("label",data['0']['labels'])
        
    print("time",data['0']['times'][0])





#!/usr/bin/env python
#-----------------------------------------------------------------------------
# Author: Emre Neftci
#
# Creation Date : Fri 01 Dec 2017 10:05:17 PM PST
# Last Modified : Sun 29 Jul 2018 01:39:06 PM PDT
#
# Copyright : (c)
# Licence : Apache License, Version 2.0
#-----------------------------------------------------------------------------

import struct
import time, copy
import numpy as np
import scipy.misc
import h5py
import torch.utils.data
from ..neuromorphic_dataset import NeuromorphicDataset 
from ..events_timeslices import *
from ..transforms import *
from .create_hdf5_sign import create_events_hdf5
import os

NUM_CLASSES = 24 # A-Y excluding j

mapping = { 'a':0,
            'b':1,
            'c':2,
            'd':3,
            'e':4,
            'f':5,
            'g':6,
            'h':7,
            'i':8,
            'k':9,
            'l':10,
            'm':11,
            'n':12,
            'o':13,
            'p':14,
            'q':15,
            'r':16,
            's':17,
            't':18,
            'u':19,
            'v':20,
            'w':21,
            'x':22,
            'y':23}

class DVSSignDataset(NeuromorphicDataset):
    resources_url = [['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAB0jgWqXDn3sZB_YXEjZLv4a/Yin%20Bi%20-%20a.zip?dl=0',None, 'Yin Bi - a.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAC-671H-Z7XTAQcT7GJXFsGa/Yin%20Bi%20-%20b.zip?dl=0', None, 'Yin Bi - b.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AADa8hkEbpgnNbBtmRIuAw3ha/Yin%20Bi%20-%20c.zip?dl=0', None, 'Yin Bi - c.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AACUrdhDl_tYnNkb8OpAJ5k4a/Yin%20Bi%20-%20d.zip?dl=0', None, 'Yin Bi - d.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AABmXnWZ2hI2dPQYn3FOClnba/Yin%20Bi%20-%20e.zip?dl=0', None, 'Yin Bi - e.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAAzopIcHXTmjPuYomjAiPtfa/Yin%20Bi%20-%20f.zip?dl=0', None, 'Yin Bi - f.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAB0PMA-VwZMM1PpJXg6q4efa/Yin%20Bi%20-%20g.zip?dl=0', None, 'Yin Bi - g.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAD9-W_I4n3lTSCMgFGgJ5Dra/Yin%20Bi%20-%20h.zip?dl=0', None, 'Yin Bi - h.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AADCm3yMnJcwGK70bYk-ycF0a/Yin%20Bi%20-%20i.zip?dl=0', None, 'Yin Bi - i.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAChe4QTo2DduVOVuT5hN9fxa/Yin%20Bi%20-%20k.zip?dl=0', None, 'Yin Bi - k.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAAeUazC7PHK85V6wkEub1iMa/Yin%20Bi%20-%20l.zip?dl=0', None, 'Yin Bi - l.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AADrzW_ts5UxlulXaItiNuCSa/Yin%20Bi%20-%20m.zip?dl=0', None, 'Yin Bi - m.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AABlYTbweHA22nA6PFujaBKFa/Yin%20Bi%20-%20n.zip?dl=0', None, 'Yin Bi - n.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AABzMSVYuZ0hb5FsUoHB53xBa/Yin%20Bi%20-%20o.zip?dl=0', None, 'Yin Bi - o.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AACOJ3z96KaLAMxgNLP1eOwga/Yin%20Bi%20-%20p.zip?dl=0', None, 'Yin Bi - p.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAC6B6UajjSuf6aYYOfOY3I7a/Yin%20Bi%20-%20q.zip?dl=0', None, 'Yin Bi - q.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAC-Gq6qzv0yiAnvozEozqoaa/Yin%20Bi%20-%20r.zip?dl=0', None, 'Yin Bi - r.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AABDbuK8B0Mferpf0x3xbDJQa/Yin%20Bi%20-%20s.zip?dl=0', None, 'Yin Bi - s.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAAEgawjqMHuY_TvqCYNC-uIa/Yin%20Bi%20-%20t.zip?dl=0', None, 'Yin Bi - t.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AABTIlScbaSqFIahMy_NRNUna/Yin%20Bi%20-%20u.zip?dl=0', None, 'Yin Bi - u.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAAp8JvzLPfhGlG5jL943W_pa/Yin%20Bi%20-%20v.zip?dl=0', None, 'Yin Bi - v.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AAA4qvsatKEDeoykc2I4a6FRa/Yin%20Bi%20-%20w.zip?dl=0', None, 'Yin Bi - w.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AABfnDu7rtZve1w9VVQQwuFia/Yin%20Bi%20-%20x.zip?dl=0', None, 'Yin Bi - x.zip'],
                     ['https://www.dropbox.com/sh/ibq0jsicatn7l6r/AADKEQBAXFQ9P0GoGHTY4ig8a/Yin%20Bi%20-%20y.zip?dl=0', None, 'Yin Bi - y.zip']
                    ]
    directory = 'data/ASL-DVS/'#'data/nmnist/'
    resources_local = [directory+'a', directory+'b', directory+'c', directory+'d',directory+'e',directory+'f',directory+'g',directory+'h',directory+'i',
                      directory+'k',directory+'l',directory+'m',directory+'n',directory+'o',directory+'p',directory+'q',directory+'r',directory+'s',
                      directory+'t',directory+'u',directory+'v',directory+'w',directory+'x',directory+'y',directory+'z']

    def __init__(
            self, 
            root,
            train=True,
            transform=None,
            target_transform=None,
            download_and_create=True,
            chunk_size = 100):

        self.n = 0
        self.nclasses = self.num_classes = 10
        self.download_and_create = download_and_create
        self.root = root
        self.train = train 
        self.chunk_size = chunk_size

        super(DVSSignDataset, self).__init__(
                root,
                transform=transform,
                target_transform=target_transform )
        
        with h5py.File(root, 'r', swmr=True, libver="latest") as f:
            try:
                if train:
                    self.n = f['extra'].attrs['Ntrain']
                    self.keys = f['extra']['train_keys'][()]
                    self.keys_by_label = f['extra']['train_keys_by_label'][()]
                else:
                    self.n = f['extra'].attrs['Ntest']
                    self.keys = f['extra']['test_keys'][()]
                    self.keys_by_label = f['extra']['test_keys_by_label'][()]
                    self.keys_by_label[:,:] -= self.keys_by_label[0,0] #normalize
            except AttributeError:
                print('Attribute not found in hdf5 file. You may be using an old hdf5 build. Delete {0} and run again'.format(root))
                raise


    def download(self):
        isexisting = super(DVSSignDataset, self).download()

    def create_hdf5(self):
        create_events_hdf5(self.directory, self.root)


    def __len__(self):
        return self.n
        
    def __getitem__(self, key):
        #Important to open and close in getitem to enable num_workers>0
        with h5py.File(self.root, 'r', swmr=True, libver="latest") as f:
            if self.train:
                key = f['extra']['train_keys'][key]
            else:
                key = f['extra']['test_keys'][key]
            data, target = sample(
                    f,
                    key,
                    T = self.chunk_size)

        if self.transform is not None:
            data = self.transform(data)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return data, target

def sample(hdf5_file,
        key,
        T = 300):
    dset = hdf5_file['data'][str(key)]
    label = dset['labels'][()]
    tend = dset['times'][-1] 
    start_time = 0

    tmad = get_tmad_slice(dset['times'][()], dset['addrs'][()], start_time, T*1000)
    tmad[:,0]-=tmad[0,0]
    return tmad, label

def create_datasets(
        root = 'data/dvssign/dvssign.hdf5',
        batch_size = 72 ,
        chunk_size_train = 300,
        chunk_size_test = 300,
        ds = 1,
        dt = 1000,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None):

    size = [2, 240//ds, 180//ds]

    if transform_train is None:
        transform_train = Compose([
            CropDims(low_crop=[0,0], high_crop=[240,180], dims=[2,3]),
            Downsample(factor=[dt,1,1,1]),
            ToCountFrame(T = chunk_size_train, size = size),
            ToTensor()])
    if transform_test is None:
        transform_test = Compose([
            CropDims(low_crop=[0,0], high_crop=[240,180], dims=[2,3]),
            Downsample(factor=[dt,1,1,1]),
            ToCountFrame(T = chunk_size_test, size = size),
            ToTensor()])
    if target_transform_train is None:
        target_transform_train =Compose([Repeat(chunk_size_train), toOneHot(NUM_CLASSES)])
    if target_transform_test is None:
        target_transform_test = Compose([Repeat(chunk_size_test), toOneHot(NUM_CLASSES)])

    train_ds = DVSSignDataset(root,train=True,
                                 transform = transform_train, 
                                 target_transform = target_transform_train, 
                                 chunk_size = chunk_size_train)

    test_ds = DVSSignDataset(root, transform = transform_test, 
                                 target_transform = target_transform_test, 
                                 train=False,
                                 chunk_size = chunk_size_test)

    return train_ds, test_ds

def create_dataloader(
        root = 'data/dvssign/dvssign.hdf5',
        batch_size = 72 ,
        chunk_size_train = 100,
        chunk_size_test = 100,
        ds = 1,
        dt = 1000,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None,
        **dl_kwargs):

    train_d, test_d = create_datasets(
        root = root,
        batch_size = batch_size,
        chunk_size_train = chunk_size_train,
        chunk_size_test = chunk_size_test,
        ds = ds,
        dt = dt,
        transform_train = transform_train,
        transform_test = transform_test,
        target_transform_train = target_transform_train,
        target_transform_test = target_transform_test)


    train_dl = torch.utils.data.DataLoader(train_d, shuffle=True, batch_size=batch_size, **dl_kwargs)
    test_dl = torch.utils.data.DataLoader(test_d, shuffle=False, batch_size=batch_size, **dl_kwargs)

    return train_dl, test_dl






#!/usr/bin/env python
#-----------------------------------------------------------------------------
# Author: Emre Neftci
#
# Creation Date : Fri 01 Dec 2017 10:05:17 PM PST
# Last Modified : Sun 29 Jul 2018 01:39:06 PM PDT
#
# Copyright : (c)
# Licence : Apache License, Version 2.0
#-----------------------------------------------------------------------------
import struct
import time, copy
import numpy as np
import scipy.misc
import h5py
import torch.utils.data
from ..nmnist.nmnist_dataloaders import NMNISTDataset, sample, create_datasets
from ..neuromorphic_dataset import NeuromorphicDataset 
from ..events_timeslices import *
from ..transforms import *
import os
import torchmeta
from torchmeta.transforms import Categorical

mapping = { 0 :'0',
            1 :'1',
            2 :'2',
            3 :'3',
            4 :'4',
            5 :'5',
            6 :'6',
            7 :'7',
            8 :'8',
            9 :'9'}

# Splits used for the double MNIST task https://github.com/shaohua0116/MultiDigitMNIST
splits = {}
splits['train'] = ['00', '01', '04', '05', '06', '08', '09', '11', '12', '13', '14', '15', '16', '18', '19', '20', '21', '23', '24', '26', '28', '29', '30', '31', '33', '35', '37', '38', '41', '42', '43', '44', '45', '50', '51', '53', '54', '56', '59', '60', '62', '63', '65', '69', '70', '72', '74', '75', '76', '77', '79', '81', '82', '84', '85', '87', '88', '89', '90', '91', '94', '95', '97', '98']
splits['val'] = ['03', '07', '10', '22', '27', '34', '39', '40', '48', '52', '58', '61', '64', '71', '93', '99']
splits['test'] = ['02', '17', '25', '32', '36', '46', '47', '49', '55', '57', '66', '67', '68', '73', '78', '80', '83', '86', '92', '96']

class DoubleNMNISTClassDataset(NeuromorphicDataset):
    def __init__(
            self, 
            root : str,
            train : bool = True,
            transform : object = None ,
            target_transform=None,
            download_and_create=True,
            chunk_size = 500,
            samples_per_class = 1,
            label_u = 0):

        self.transform = transform
        self.target_transform = target_transform
        
        self.samples_per_class = samples_per_class
        self.download_and_create = download_and_create
        self.root = root
        self.train = train 

        self.chunk_size = chunk_size
        self.label_u = label_u


        lu = self.label_u
        self.labels_left = lu // 10
        self.labels_right = lu % 10

        ll = self.labels_left
        lr = self.labels_right

        #self.labels.append( np.repeat(self.labels_u, self.samples_per_class))
        #self.labels_map.append( dict(zip(np.unique(self.labels[i]),np.arange(self.nclasses))))

        super(DoubleNMNISTClassDataset, self).__init__(root = None,
                                            transform=transform,
                                            target_transform=target_transform )
       
        self.data_orig = NMNISTDataset( root,
                      train=train,
                      transform=transform,
                      target_transform=None,
                      download_and_create=download_and_create,                    
                      chunk_size = chunk_size)

        self.nl = len(self.data_orig.keys_by_label[ll])
        self.nr = len(self.data_orig.keys_by_label[lr])
        self.n = self.nl * self.nr


    def __len__(self):
        return 1000 #self.n

    def __getitem__(self, key):
        ll = self.labels_left
        lr = self.labels_right
        key_l = self.data_orig.keys_by_label[ll][ key // self.nl] 
        key_r = self.data_orig.keys_by_label[lr][ key % self.nl ] 

        data_l, label_l =  self.data_orig[key_l]
        data_r, label_r =  self.data_orig[key_r]

        size_x, size_y = data_r.shape[2:4]
        data = torch.zeros(data_r.shape[:2]+(size_x*2,size_y*2))
        np.random.seed(key%1313)
        r1 = np.random.randint(0,size_y)
        r2 = np.random.randint(0,size_y)
        data[:, :, :size_x, r1:r1+size_y] = data_l
        data[:, :, size_x:, r2:r2+size_y] = data_r
        target = self.label_u
        return data, self.target_transform(target)




class ClassNMNISTDataset(torchmeta.utils.data.ClassDataset):
    def __init__(self, root = 'data/nmnist/n_mnist.hdf5', chunk_size=300, meta_train=False, meta_val=False, meta_test=False, meta_split=None, transform=None, target_transform=None, download=False, class_augmentations=None):
        self.root=root
        self.chunk_size = chunk_size
        if meta_train is True:
            train = True
        else:
            train = False
        
        if meta_train:
            split_name = 'train'
        if meta_val:
            split_name = 'val'
        if meta_test:
            split_name = 'test'
        self.split_name = split_name

        self.transform = transform


        super(ClassNMNISTDataset, self).__init__(
                meta_train=meta_train,
                meta_val=meta_val,
                meta_test=meta_test,
                meta_split=meta_split,
                class_augmentations=class_augmentations)


        self._labels = [int(s) for s in splits[split_name]]
        self._num_classes = len(self._labels)

    @property
    def labels(self):
        return self._labels

    @property
    def num_classes(self):
        return self._num_classes

    def __getitem__(self, index):
        label = self._labels[index]
        d = DoubleNMNISTClassDataset(root =self.root, train= self.meta_train, label_u = label, transform = self.transform, target_transform = None, chunk_size = self.chunk_size)
        d.index = index
        #d.target_transform_append = lambda x: None
        return d

def create_class_dataset(dset, meta_split = 'train'):
    ds = []
    for n in range(dset.nclasses):
        indices = dset.keys_by_label[n]
        d = torch.utils.data.Subset(dset, indices)
        ds.append(d)
    return ds

class DoubleNMNIST(torchmeta.utils.data.CombinationMetaDataset):
    def __init__(self, root, num_classes_per_task=None, meta_train=False,
                 meta_val=False, meta_test=False, meta_split=None,
                 transform=None, target_transform=None, dataset_transform=None,
                 class_augmentations=None, download=False,chunk_size=300):

        target_transform = Categorical(num_classes_per_task)
        dataset = ClassNMNISTDataset(root,
            meta_train=meta_train, meta_val=meta_val,
            meta_test=meta_test, meta_split=meta_split, transform=transform,
            class_augmentations=class_augmentations, download=download,chunk_size=300)

        super(DoubleNMNIST, self).__init__(dataset, num_classes_per_task,
            target_transform=target_transform,
            dataset_transform=dataset_transform)





#!/usr/bin/env python
#-----------------------------------------------------------------------------
# Author: Emre Neftci
#
# Creation Date : Fri 01 Dec 2017 10:05:17 PM PST
# Last Modified : Sun 29 Jul 2018 01:39:06 PM PDT
#
# Copyright : (c)
# Licence : Apache License, Version 2.0
#-----------------------------------------------------------------------------
import struct
import time
import numpy as np
import scipy.misc
import h5py
import torch.utils.data
from .create_hdf5 import create_events_hdf5
from ..neuromorphic_dataset import NeuromorphicDataset 
from ..events_timeslices import *
from ..transforms import *
import os

mapping = { 0 :'Hand Clapping'  ,
            1 :'Right Hand Wave',
            2 :'Left Hand Wave' ,
            3 :'Right Arm CW'   ,
            4 :'Right Arm CCW'  ,
            5 :'Left Arm CW'    ,
            6 :'Left Arm CCW'   ,
            7 :'Arm Roll'       ,
            8 :'Air Drums'      ,
            9 :'Air Guitar'     ,
            10:'Other'}

class DVSGestureDataset(NeuromorphicDataset):
    directory = 'data/dvsgesture/'
    resources_url = [['Manually Download dataset here: https://ibm.ent.box.com/s/3hiq58ww1pbbjrinh367ykfdf60xsfm8/file/211521748942?sb=/details and place under {0}'.format(directory),None, 'DvsGesture.tar.gz']]
    resources_local = [directory+'raw']

    def __init__(
            self, 
            root,
            train=True,
            transform=None,
            target_transform=None,
            download_and_create=True,
            chunk_size = 500,
            return_meta = False):

        self.n = 0
        self.download_and_create = download_and_create
        self.root = root
        self.train = train 
        self.chunk_size = chunk_size
        self.return_meta = return_meta

        super(DVSGestureDataset, self).__init__(
                root,
                transform=transform,
                target_transform=target_transform )

        with h5py.File(root, 'r', swmr=True, libver="latest") as f:
            if train:
                self.n = f['extra'].attrs['Ntrain']
                self.keys = f['extra']['train_keys'][()]
            else:
                self.n = f['extra'].attrs['Ntest']
                self.keys = f['extra']['test_keys'][()]

    def download(self):
        super(DVSGestureDataset, self).download()

    def create_hdf5(self):
        create_events_hdf5(self.resources_local[0], self.root)

    def __len__(self):
        return self.n
        
    def __getitem__(self, key):
        #Important to open and close in getitem to enable num_workers>0
        with h5py.File(self.root, 'r', swmr=True, libver="latest") as f:
            if not self.train:
                key = key + f['extra'].attrs['Ntrain']
            assert key in self.keys
            data, target, meta_info = sample(
                    f,
                    key,
                    T = self.chunk_size,
                    shuffle=self.train)

        if self.transform is not None:
            data = self.transform(data)

        if self.target_transform is not None:
            target = self.target_transform(target)

        if self.return_meta is True:
            return data, target, meta_info
        else:
            return data, target

def sample(hdf5_file,
        key,
        T = 500,
        shuffle = False):
    dset = hdf5_file['data'][str(key)]
    label = dset['labels'][()]
    tbegin = dset['times'][0]
    tend = np.maximum(0,dset['times'][-1]- 2*T*1000 )
    start_time = np.random.randint(tbegin, tend) if shuffle else 0

    tmad = get_tmad_slice(dset['times'][()], dset['addrs'][()], start_time, T*1000)
    tmad[:,0]-=tmad[0,0]
    return tmad[:, [0,3,1,2]], label, dset.attrs['meta_info']

def create_dataloader(
        root = 'data/dvsgesture/dvs_gestures_build19.hdf5',
        batch_size = 72 ,
        chunk_size_train = 500,
        chunk_size_test = 1800,
        ds = None,
        dt = 1000,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None,
        n_events_attention=None,
        return_meta=False,
        **dl_kwargs):
    if ds is None:
        ds = 4
    if isinstance(ds,int):
        ds = [ds,ds]
        
    size = [2, 128//ds[0], 128//ds[1]]

    if n_events_attention is None:
        default_transform = lambda chunk_size: Compose([
            Downsample(factor=[dt,1,ds[0],ds[1]]),
            ToCountFrame(T = chunk_size, size = size),
            ToTensor()
        ])
    else:
        default_transform = lambda chunk_size: Compose([
            Downsample(factor=[dt,1,1,1]),
            Attention(n_events_attention, size=size),
            ToCountFrame(T = chunk_size, size = size),
            ToTensor()
        ])

    if transform_train is None:
        transform_train = default_transform(chunk_size_train)
    if transform_test is None:
        transform_test = default_transform(chunk_size_test)

    if target_transform_train is None:
        target_transform_train =Compose([Repeat(chunk_size_train), toOneHot(11)])
    if target_transform_test is None:
        target_transform_test = Compose([Repeat(chunk_size_test), toOneHot(11)])

    train_d = DVSGestureDataset(root,
                                train=True,
                                transform = transform_train, 
                                target_transform = target_transform_train, 
                                chunk_size = chunk_size_train,
                                return_meta = return_meta)

    train_dl = torch.utils.data.DataLoader(train_d, batch_size=batch_size, shuffle=True, **dl_kwargs)

    test_d = DVSGestureDataset(root,
                               transform = transform_test, 
                               target_transform = target_transform_test, 
                               train=False,
                               chunk_size = chunk_size_test,
                               return_meta = return_meta)

    test_dl = torch.utils.data.DataLoader(test_d, batch_size=batch_size, **dl_kwargs)

    return train_dl, test_dl






#!/bin/python
#-----------------------------------------------------------------------------
# File Name : create_hdf5.py
# Author: Emre Neftci
#
# Creation Date : Tue Nov  5 13:15:54 2019
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv2
#----------------------------------------------------------------------------- 
import numpy as np
from tqdm import tqdm
import scipy.misc
import h5py
import glob
import torch.utils.data
from ..events_timeslices import *
from ..utils import *
import os

from collections import namedtuple, defaultdict
import torch
import torch.utils.data
from ..utils import load_ATIS_bin, load_jaer

def nmnist_load_events_from_bin(file_path, max_duration=None):
    timestamps, xaddr, yaddr, pol = load_ATIS_bin(file_path)
    return np.column_stack([
        np.array(timestamps, dtype=np.uint32),
        np.array(pol, dtype=np.uint8),
        np.array(xaddr, dtype=np.uint16),
        np.array(yaddr, dtype=np.uint16)])

def nmnist_get_file_names(dataset_path):
    if not os.path.isdir(dataset_path):
        raise FileNotFoundError("N-MNIST Dataset not found, looked at: {}".format(dataset_path))

    train_files = []
    test_files = []
    for digit in range(10):
        digit_train = glob.glob(os.path.join(dataset_path, 'Train/{}/*.bin'.format(digit)))
        digit_test = glob.glob(os.path.join(dataset_path, 'Test/{}/*.bin'.format(digit)))
        train_files.append(digit_train)
        test_files.append(digit_test)

    # We need the same number of train and test samples for each digit, let's compute the minimum
    max_n_train = min(map(lambda l: len(l), train_files))
    max_n_test = min(map(lambda l: len(l), test_files))
    n_train = max_n_train # we could take max_n_train, but my memory on the shared drive is full
    n_test = max_n_test # we test on the whole test set - lets only take 100*10 samples
    assert((n_train <= max_n_train) and (n_test <= max_n_test)), 'Requested more samples than present in dataset'

    print("N-MNIST: {} train samples and {} test samples per digit (max: {} train and {} test)".format(n_train, n_test, max_n_train, max_n_test))
    # Crop extra samples of each digits
    train_files = map(lambda l: l[:n_train], train_files)
    test_files = map(lambda l: l[:n_test], test_files)

    return list(train_files), list(test_files)

def create_events_hdf5(directory, hdf5_filename):
    fns_train, fns_test = nmnist_get_file_names(directory)
    fns_train = [val for sublist in fns_train for val in sublist]
    fns_test = [val for sublist in fns_test for val in sublist]
    test_keys = []
    train_keys = []
    train_label_list = [[] for i in range(10)]
    test_label_list = [[] for i in range(10)]

    with h5py.File(hdf5_filename, 'w') as f:
        f.clear()
        key = 0
        metas = []
        data_grp = f.create_group('data')
        extra_grp = f.create_group('extra')
        for file_d in tqdm(fns_train+fns_test):
            istrain = file_d in fns_train
            data = nmnist_load_events_from_bin(file_d)
            times = data[:,0]
            addrs = data[:,1:]
            label = int(file_d.split('/')[-2])
            out = []

            if istrain: 
                train_keys.append(key)
                train_label_list[label].append(key)
            else:
                test_keys.append(key)
                test_label_list[label].append(key)
            metas.append({'key':str(key), 'training sample':istrain}) 
            subgrp = data_grp.create_group(str(key))
            tm_dset = subgrp.create_dataset('times' , data=times, dtype = np.uint32)
            ad_dset = subgrp.create_dataset('addrs' , data=addrs, dtype = np.uint8)
            lbl_dset= subgrp.create_dataset('labels', data=label, dtype = np.uint8)
            subgrp.attrs['meta_info']= str(metas[-1])
            assert label in range(10)
            key += 1
        extra_grp.create_dataset('train_keys', data = train_keys)
        extra_grp.create_dataset('train_keys_by_label', data = train_label_list)
        extra_grp.create_dataset('test_keys_by_label', data = test_label_list)
        extra_grp.create_dataset('test_keys', data = test_keys)
        extra_grp.attrs['N'] = len(train_keys) + len(test_keys)
        extra_grp.attrs['Ntrain'] = len(train_keys)
        extra_grp.attrs['Ntest'] = len(test_keys)



#!/usr/bin/env python
#-----------------------------------------------------------------------------
# Author: Emre Neftci
#
# Creation Date : Fri 01 Dec 2017 10:05:17 PM PST
# Last Modified : Sun 29 Jul 2018 01:39:06 PM PDT
#
# Copyright : (c)
# Licence : Apache License, Version 2.0
#-----------------------------------------------------------------------------
import struct
import time, copy
import numpy as np
import scipy.misc
import h5py
import torch.utils.data
from ..nmnist.nmnist_dataloaders import NMNISTDataset, sample, create_datasets
from ..neuromorphic_dataset import NeuromorphicDataset 
from ..events_timeslices import *
from ..transforms import *
import os

mapping = { 0 :'0',
            1 :'1',
            2 :'2',
            3 :'3',
            4 :'4',
            5 :'5',
            6 :'6',
            7 :'7',
            8 :'8',
            9 :'9'}

class DoubleNMNISTDataset(NeuromorphicDataset):

    def __init__(
            self, 
            root,
            train=True,
            transform=None,
            target_transform=None,
            download_and_create=True,
            chunk_size = 500,
            nclasses = 5,
            samples_per_class = 2,
            labels_u = range(5)):

        self.n = samples_per_class * nclasses
        self.nclasses = nclasses
        self.download_and_create = download_and_create
        self.root = root
        self.train = train 
        self.chunk_size = chunk_size
        self.labels_u = labels_u
        labels_left =  self.labels_u // 10
        labels_right =  self.labels_u % 10
        self.labels = np.repeat(self.labels_u, samples_per_class)
        self.labels_map =  dict(zip(np.unique(self.labels),np.arange(nclasses)))


        super(DoubleNMNISTDataset, self).__init__(root = None,
                                            transform=transform,
                                            target_transform=target_transform )
       
        self.data_train = NMNISTDataset( root,
                      train=True,
                      transform=transform,
                      target_transform=target_transform,
                      download_and_create=download_and_create,                    
                      chunk_size = chunk_size)


        keys_filt_left  = np.array([ np.random.choice(s, samples_per_class) for s in self.data_train.keys_by_label[labels_left]]).reshape(-1)
        keys_filt_right = np.array([ np.random.choice(s, samples_per_class) for s in self.data_train.keys_by_label[labels_right]]).reshape(-1)
        self.keys = list(zip(keys_filt_left, keys_filt_right))

    def __len__(self):
        return self.n
        
    def __getitem__(self, key):
        key_l, key_r = self.keys[key]
        data_l, label_l =  self.data_train[key_l]
        data_r, label_r =  self.data_train[key_r]
        size_x, size_y = data_r.shape[2:4]
        data = torch.zeros(data_r.shape[:2]+(size_x*2,size_y*2))
        r1 = np.random.randint(0,size_y)
        r2 = np.random.randint(0,size_y)
        data[:, :, :size_x, r1:r1+size_y] = data_l
        data[:, :, size_x:, r2:r2+size_y] = data_r
        target = self.labels_map[self.labels[key]]
        return data, target

def create_datasets(
        root = 'data/nmnist/n_mnist.hdf5',
        batch_size = 72 ,
        chunk_size_train = 300,
        chunk_size_test = 300,
        ds = 1,
        dt = 1000,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None,
        nclasses = 5,
        samples_per_class = 2,
        samples_per_test = 2,
        classes_meta = np.arange(100, dtype='int')):

    size = [2, 32//ds, 32//ds]

    if transform_train is None:
        transform_train = Compose([
            CropDims(low_crop=[0,0], high_crop=[32,32], dims=[2,3]),
            Downsample(factor=[dt,1,ds,ds]),
            ToCountFrame(T = chunk_size_train, size = size),
            ToTensor()])
    if transform_test is None:
        transform_test = Compose([
            CropDims(low_crop=[0,0], high_crop=[32,32], dims=[2,3]),
            Downsample(factor=[dt,1,ds,ds]),
            ToCountFrame(T = chunk_size_test, size = size),
            ToTensor()])
    if target_transform_train is None:
        target_transform_train =Compose([Repeat(chunk_size_train), toOneHot(nclasses)])
    if target_transform_test is None:
        target_transform_test = Compose([Repeat(chunk_size_test), toOneHot(nclasses)])


    labels_u = np.random.choice(classes_meta, nclasses,replace=False) #100 here becuase we have two pairs of digits between 0 and 9

    train_ds = DoubleNMNISTDataset(root,train=True,
                                 transform = transform_train, 
                                 target_transform = target_transform_train, 
                                 chunk_size = chunk_size_train,
                                 nclasses = nclasses,
                                 samples_per_class = samples_per_class,
                                 labels_u = labels_u)

    test_ds = DoubleNMNISTDataset(root, transform = transform_test, 
                                 target_transform = target_transform_test, 
                                 train=False,
                                 chunk_size = chunk_size_test,
                                 nclasses = nclasses,
                                 samples_per_class = samples_per_test,
                                 labels_u = labels_u)

    return train_ds, test_ds


def create_dataloader(
        root = 'data/nmnist/n_mnist.hdf5',
        batch_size = 72 ,
        chunk_size_train = 300,
        chunk_size_test = 300,
        ds = 1,
        dt = 1000,
        transform_train = None,
        transform_test = None,
        target_transform_train = None,
        target_transform_test = None,
        nclasses = 5,
        samples_per_class = 2,
        samples_per_test = 2,
        classes_meta = np.arange(100, dtype='int'),
        **dl_kwargs):


    train_d, test_d = create_datasets(
        root = root,
        batch_size = batch_size,
        chunk_size_train = chunk_size_train,
        chunk_size_test = chunk_size_test,
        ds = ds,
        dt = dt,
        transform_train = transform_train,
        transform_test = transform_test,
        target_transform_train = target_transform_train,
        target_transform_test = target_transform_test,
        classes_meta = classes_meta,
        nclasses = nclasses,
        samples_per_class = samples_per_class,
        samples_per_test = samples_per_test)


    train_dl = torch.utils.data.DataLoader(train_d, shuffle=True, batch_size=batch_size, **dl_kwargs)
    test_dl = torch.utils.data.DataLoader(test_d, shuffle=False, batch_size=batch_size, **dl_kwargs)

    return train_dl, test_dl




def sample_double_mnist_task( N = 5,
                              K = 2,
                              K_test = 2,
                              meta_split = [range(64), range(64,80), range(80,100)],
                              meta_dataset_type = 'train',
                              **kwargs):
    classes_meta = {}
    classes_meta['train'] = np.array(meta_split[0], dtype='int')
    classes_meta['val']   = np.array(meta_split[1], dtype='int')
    classes_meta['test']  = np.array(meta_split[2], dtype='int')

    assert meta_dataset_type in ['train', 'val', 'test']
    return create_dataloader(classes_meta = classes_meta[meta_dataset_type], nclasses= N, samples_per_class = K, samples_per_test = K_test, **kwargs)



#!/bin/python
#-----------------------------------------------------------------------------
# File Name : create_hdf5.py
# Author: Emre Neftci
#
# Creation Date : Tue Nov  5 13:15:54 2019
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv2
#----------------------------------------------------------------------------- 
import numpy as np
from tqdm import tqdm
import scipy.misc
import h5py
import glob
import torch.utils.data
from ..events_timeslices import *
from ..utils import *
import os

from collections import namedtuple, defaultdict
import torch
import torch.utils.data
from ..utils import load_ATIS_bin, load_jaer

NUM_CLASSES = 24 # A-Y excluding j

mapping = { 'a':0,
            'b':1,
            'c':2,
            'd':3,
            'e':4,
            'f':5,
            'g':6,
            'h':7,
            'i':8,
            'k':9,
            'l':10,
            'm':11,
            'n':12,
            'o':13,
            'p':14,
            'q':15,
            'r':16,
            's':17,
            't':18,
            'u':19,
            'v':20,
            'w':21,
            'x':22,
            'y':23}

def sign_load_events_from_mat(file_path, max_duration=None):
    timestamps, xaddr, yaddr, pol = load_mat(file_path)#load_ATIS_bin(file_path)
    return np.column_stack([
        np.array(timestamps, dtype=np.uint32),
        np.array(pol, dtype=np.uint8),
        np.array(xaddr, dtype=np.uint16), 
        np.array(yaddr, dtype=np.uint16)])

def sign_get_file_names(dataset_path):
    if not os.path.isdir(dataset_path):
        raise FileNotFoundError("DVSSign Dataset not found, looked at: {}".format(dataset_path))

    sign_dict_train = {}
    
    sign_dict_test = {}
    
    # need to create train test split
    # simple 80/20 split based on number of gestures per class on a fraction of the dataset
    num_train = 400 #3360
    num_test = 100 #840
    os.chdir(dataset_path+'/a')
    for key in mapping.keys():
        os.chdir('../'+key)
        curr_dir = os.getcwd()
        # each class has 4200 samples.
        # if doing 80/20 split, 3360 train 840 test
        num_samples = 1
        for file in glob.glob("*.mat"):
            label = file.split('_')[0]

            if num_samples < num_train:
                if label in sign_dict_train.keys():
                    sign_dict_train[label].append(curr_dir+'/'+file)
                else:
                    sign_dict_train[label] = []
                    sign_dict_train[label].append(curr_dir+'/'+file)
            if num_samples < num_train+num_test:
                if label in sign_dict_test.keys():
                    sign_dict_test[label].append(curr_dir+'/'+file)
                else:
                    sign_dict_test[label] = []
                    sign_dict_test[label].append(curr_dir+'/'+file)
            num_samples += 1

    # We need the same number of train and test samples for each digit, let's compute the minimum
    #max_n_train = min(map(lambda l: len(l), train_files))
    #max_n_test = min(map(lambda l: len(l), test_files))
    #n_train = max_n_train # we could take max_n_train, but my memory on the shared drive is full
    #n_test = max_n_test # we test on the whole test set - lets only take 100*10 samples
    #assert((n_train <= max_n_train) and (n_test <= max_n_test)), 'Requested more samples than present in dataset'

    print("DVSSign: {} train samples and {} test samples per digit (max: {} train and {} test)".format(num_train*NUM_CLASSES, num_test*NUM_CLASSES, num_train*NUM_CLASSES, num_test*NUM_CLASSES))
    # Crop extra samples of each digits
    #train_files = map(lambda l: l[:n_train], train_files)
    #test_files = map(lambda l: l[:n_test], test_files)

    return sign_dict_train, sign_dict_test

def create_events_hdf5(directory, hdf5_filename):
    fns_train, fns_test = sign_get_file_names(directory)
    #fns_train = [val for sublist in fns_train for val in sublist]
    #fns_test = [val for sublist in fns_test for val in sublist]
    test_keys = []
    train_keys = []
    train_label_list = [[] for i in range(NUM_CLASSES)]
    test_label_list = [[] for i in range(NUM_CLASSES)]

    with h5py.File(hdf5_filename, 'w') as f:
        f.clear()
        #key = 0
        num = 0
        metas = []
        data_grp = f.create_group('data')
        extra_grp = f.create_group('extra')
        for key in tqdm(fns_train.keys()):
            for file_d in fns_train.values():
                for i in range(len(file_d)):
                    data = sign_load_events_from_mat(file_d[i])
                    times = data[:,0]
                    addrs = data[:,1:]
                    label = mapping[key] #int(file_d.split('/')[-2]) 

                    train_keys.append(num)

                    train_label_list[mapping[key]].append(num)

                    metas.append({'key':str(num), 'training sample':True}) 
                    subgrp = data_grp.create_group(str(num))
                    tm_dset = subgrp.create_dataset('times' , data=times, dtype = np.uint32)
                    ad_dset = subgrp.create_dataset('addrs' , data=addrs, dtype = np.uint8)
                    lbl_dset= subgrp.create_dataset('labels', data=label, dtype = np.uint8)
                    subgrp.attrs['meta_info']= str(metas[-1])
                    num += 1
                
            for file_d in fns_test.values():
                for i in range(len(file_d)):
                    data = sign_load_events_from_mat(file_d[i])
                    times = data[:,0]
                    addrs = data[:,1:]
                    label = mapping[key]

                    test_keys.append(num)

                    test_label_list[mapping[key]].append(num)

                    metas.append({'key':str(num), 'testing sample':True}) 
                    subgrp = data_grp.create_group(str(num))
                    tm_dset = subgrp.create_dataset('times' , data=times, dtype = np.uint32)
                    ad_dset = subgrp.create_dataset('addrs' , data=addrs, dtype = np.uint8)
                    lbl_dset= subgrp.create_dataset('labels', data=label, dtype = np.uint8)
                    subgrp.attrs['meta_info']= str(metas[-1])
                    num += 1
                
        extra_grp.create_dataset('train_keys', data = train_keys)
        extra_grp.create_dataset('train_keys_by_label', data = train_label_list)
        extra_grp.create_dataset('test_keys_by_label', data = test_label_list)
        extra_grp.create_dataset('test_keys', data = test_keys)
        extra_grp.attrs['N'] = len(train_keys) + len(test_keys)
        extra_grp.attrs['Ntrain'] = len(train_keys)
        extra_grp.attrs['Ntest'] = len(test_keys)



#!/usr/bin/env python
#-----------------------------------------------------------------------------
# Author: Jacques Kaiser
#
# Creation Date : Wed 13 May 2020
#
# Copyright : (c)
# Licence : Apache License, Version 2.0
#-----------------------------------------------------------------------------

import numpy as np
import h5py
from ..events_timeslices import *
from ..utils import *
import torch.utils.data
import glob
import os
from tqdm import tqdm

def gather_rosbags_class_folders(directory):
    # expected folder structure:
    # directory
    # ├── test_list.csv
    # ├── label1
    # │__ ├── bag1.bag
    # │__ ├── bag2.bag
    # ├── label2
    # │__ ├── bag1.bag
    # │__ ├── bag2.bag
    all_bags = glob.glob(os.path.join(directory, '**/*.bag'))
    test_list = np.loadtxt(os.path.join(directory, 'test_list.csv'), delimiter=',',dtype=str)

    test_bags = [ bag for bag in all_bags if np.any([ bag.endswith(test) for test in test_list]) ]
    train_bags = [ bag for bag in all_bags if bag not in test_bags ]
    all_classes = np.unique([os.path.basename(os.path.dirname(b)) for b in all_bags])
    return train_bags, test_bags, all_classes


def create_events_hdf5(directory, hdf5_filename, gather_rosbags=gather_rosbags_class_folders):
    train_bags, test_bags, all_classes = gather_rosbags(directory)
    label_mapping = { label: i
                      for i,label in enumerate(all_classes)
    }
    label_mapping_inv = { i: label for label,i in label_mapping.items() }
    label_order = [label_mapping_inv[i] for i in range(len(all_classes)) ]

    with h5py.File(hdf5_filename, 'w') as f:
        f.clear()
        key = 0
        metas = []
        data_grp = f.create_group('data')
        extra_grp = f.create_group('extra')
        train_keys = []
        test_keys = []
        for file_d in train_bags + test_bags:
            filename=os.path.basename(file_d)
            class_name=os.path.basename(os.path.dirname(file_d))
            label=label_mapping[class_name]
            events=rosbag_to_events(file_d)
            addrs = np.array([events['pol'], events['x'], events['y']]).T
            istrain = file_d in train_bags
            if istrain:
                train_keys.append(key)
            else:
                test_keys.append(key)

            metas.append({'key':str(key), 'training sample':istrain})

            subgrp = data_grp.create_group(str(key))
            tm_dset = subgrp.create_dataset('times' , data=events['ts'], dtype=np.uint32)
            ad_dset = subgrp.create_dataset('addrs' , data=addrs, dtype=np.uint8)
            lbl_dset= subgrp.create_dataset('labels', data=label, dtype=np.uint8)
            subgrp.attrs['meta_info']= str(metas[-1])
            key+=1
        extra_grp.create_dataset('train_keys', data=train_keys)
        extra_grp.create_dataset('test_keys', data=test_keys)
        extra_grp.create_dataset('label_order', data=np.array(label_order, dtype='S10'))
        extra_grp.attrs['N'] = len(train_keys) + len(test_keys)
        extra_grp.attrs['Ntrain'] = len(train_keys)
        extra_grp.attrs['Ntest'] = len(test_keys)



#!/bin/python
#-----------------------------------------------------------------------------
# File Name : test_nmnist.py
# Author: Emre Neftci
#
# Creation Date : Thu Nov  7 20:30:14 2019
# Last Modified : 
#
# Copyright : (c) UC Regents, Emre Neftci
# Licence : GPLv2
#----------------------------------------------------------------------------- 
from torchneuromorphic.dvssign.dvssign_dataloaders import *
from torchneuromorphic.utils import plot_frames_imshow
import matplotlib.pyplot as plt

if __name__ == "__main__":
    train_dl, test_dl = create_dataloader(
            root='data/ASL-DVS/dvssign.hdf5',
            batch_size=8,
            ds=1,
            chunk_size_train = 100,
            chunk_size_test = 100,
            num_workers=0)
    
    iter_meta_train = iter(train_dl)
    iter_meta_test = iter(test_dl)
    frames_train, labels_train = next(iter_meta_train)
    frames_test , labels_test  = next(iter_meta_test)

    print(frames_train.shape)
    print(labels_train.shape)
    plot_frames_imshow(frames_train, labels_train, do1h=False, nim=4, avg=25)
    plt.savefig('dvssigns4.png')
